const vm = require('vm');

// filter out 'GLOBAL' and 'root' as these generate deprecation warnings
const safeContext = `
(function() {
	Function = undefined;
	const keys = Object.getOwnPropertyNames(this)
		.filter(a => a !== 'GLOBAL' && a !== 'root')
		.concat(['constructor']);
	for (const key of keys) {
		const item = this[key];
		if (key === 'data' && data && typeof item === 'object') {
			Object.defineProperty(this[key], 'constructor', {
				configurable: true
			});
		} else if (item && typeof item.constructor === 'function') {
			this[key].constructor = undefined;
		}
	}
})();
`;

async function Execute(params) {
	const { data, code, unsafe } = params;
	if (typeof code !== 'string') {
		// FIXME: Throw error instead in a breaking change
		// eslint-disable-next-line no-throw-literal
		throw 'invalid argument: code';
	}
	// determine the correct action to use
	const action = unsafe ? executeUnsafe : executeSafe;
	return action(data, code);
}

async function executeSafe(data, code) {
	// Define a new async function to resolve within a sandbox environment
	const script = new vm.Script(`(async function() { ${safeContext} ${code} })();`);
	const value = script.runInNewContext({ data, console });
	// Since we deleted them in safeContext earlier,
	// restore the constructors of the external objects
	if (data && data.constructor === undefined) {
		// While it looks strange, deleting the instance property
		// allows the non-enumerable constructor on their prototype
		// to be accessible again
		delete data.constructor;
	}
	// eslint-disable-next-line no-console
	delete console.constructor;
	return value;
}

async function executeUnsafe(data, code) {
	// Define a new async function to resolve without a sandbox environment
	// eslint-disable-next-line no-new-func
	const fn = new Function('data', 'require', `return (async function() { ${code} })();`);
	// To be functionally consistent with the safe version, make 'this' the global object
	// A number of items are not global - require, export, module, __dirname, __filename.
	// We just care about being able to use require in this function during unsafe mode.
	return fn.call(global, data, require);
}

module.exports = {
	Execute
};
